#ifndef MESH_H_FILE
#define MESH_H_FILE

#include <assert.h>
#include <float.h>
#include <string.h>
#include <stdio.h>

/**
 * Abstract base class for a mesh.  This is essentially a list of
 * vertices and a list of faces that index into the vertex list
 *
 * @author Scott Schaefer
 */

class Face
{
protected:
	int *data;
	int length;
	float r;
	float g;
	float b;

public:
	Face ( int length = 0 )
	{
		this->length = length;
		data = new int [ length ];

		r = 0.97f;
		g = 0.53f;
		b = 0.3f;

		memset ( data, 0, sizeof ( int ) * length );
	}

	Face ( const Face& f )
	{
		length = f.length;
		data = new int [ length ];
		r = f.r;
		g = f.g;
		b = f.b;

		memcpy ( data, f.data, sizeof ( int ) * length );
	}

	~Face ( void )
	{
		delete[] data;
	}

  int& operator[] ( int ind )
  {
    assert ( ind >= 0 && ind < length );
    return data [ ind ];
  }

	int *getData ( void )
	{
		return data;
	}

	int getLength ( void ) const
	{
		return length;
	}

	int operator = ( Face &f )
	{
		copy ( &f );

		return 1;
	}

	void copy ( Face *f )
	{
		if ( length != f->getLength ( ) )
		{
			delete[] data;

			length = f->getLength ( );
			data = new int [ length ];
		}

		memcpy ( data, f->getData ( ), length * sizeof ( int ) );

		this->r = f->r;
		this->g = f->g;
		this->b = f->b;
	}

	int getDimension ( void ) const
	{
		if ( length == 3 )
		{
			return 4;
		}

		return length;
	}

	void setRGB ( float r, float g, float b )
	{
		this->r = r;
		this->g = g;
		this->b = b;
	}

	float getRed ( void ) const
	{
		return r;
	}

	float getGreen ( void ) const
	{
		return g;
	}

	float getBlue ( void ) const
	{
		return b;
	}

	float *getColor ( void )
	{
		return &r;
	}
};

class Vertex
{
protected:
	float *data;
	int length;

public:
	Vertex ( int length = 8 )
	{
		this->length = length;
		data = new float [ length ];

		zero ( );
	}

	Vertex ( const Vertex &v )
	{
		length = v.length;
		data = new float [ length ];
		
		memcpy ( data, v.data, length * sizeof ( float ) );
	}

	~Vertex ( void )
	{
		delete[] data;
	}

  float& operator[] ( int ind )
  {
    assert ( ind >= 0 && ind < length );
    return data [ ind ];
  }

	int operator = ( Vertex &v )
	{
		copy ( &v );

		return 1;
	}

	void copy ( Vertex *v )
	{
		if ( v->getLength ( ) != length )
		{
			delete[] data;

			length = v->getLength ( );
			data = new float [ length ];
		}

		memcpy ( data, v->data, length * sizeof ( float ) );
	}

	void zero ( void )
	{
		int i;

		for ( i = 0; i < length; i++ )
		{
			data [ i ] = 0;
		}
	}

	float getX ( void ) const
	{
		return data [ 0 ];
	}

	float getY ( void ) const
	{
		return data [ 1 ];
	}

	float getZ ( void ) const
	{
		return data [ 2 ];
	}

	float getTexU ( void ) const
	{
		return data [ 3 ];
	}

	float getTexV ( void ) const
	{
		return data [ 4 ];
	}

	float getNormX ( void ) const
	{
		return data [ 5 ];
	}

	float getNormY ( void ) const
	{
		return data [ 6 ];
	}

	float getNormZ ( void ) const
	{
		return data [ 7 ];
	}

	void setX ( float x ) 
	{
		data [ 0 ] = x;
	}

	void setY ( float y )
	{
		data [ 1 ] = y;
	}

	void setZ ( float z )
	{
		data [ 2 ] = z;
	}

	void setTexU ( float u )
	{
		data [ 3 ] = u;
	}

	void setTexV ( float v )
	{
		data [ 4 ] = v;
	}

	void setNormX ( float nx )
	{
		data [ 5 ] = nx;
	}

	void setNormY ( float ny )
	{
		data [ 6 ] = ny;
	}

	void setNormZ ( float nz )
	{
		data [ 7 ] = nz;
	}

	float *getVert ( void )
	{
		return data;
	}

	float *getNorm ( void )
	{
		return &data [ 5 ];
	}

	float *getTex ( void )
	{
		return &data [ 3 ];
	}

	int getLength ( void ) const
	{
		return length;
	}

	void print ( void ) const
	{
		int i;

		printf ( "{" );
		for ( i = 0; i < getLength ( ); i++ )
		{
			printf ( "%.2f ", data [ i ] );
		}
		printf ( "}\n" );
	}
};


class Mesh 
{
protected:
	// number of components per vertex
	unsigned int fence;
	bool faceColors;

public:
	int numComponents;
	/**
	 * Constructor
	 *
	 * @param numComp the number of components 
	 *        (3=geometry, 5=geo+texture, 8=geo+text+normal)
	 */
	Mesh ( int numComp = 3 )
	{
		numComponents = numComp;

		faceColors = false;
	}

	virtual ~Mesh ( void )
	{
	}

	void setHasFaceColors ( bool val )
	{
		faceColors = val;
	}

	bool hasFaceColors ( void ) const
	{
		return faceColors;
	}

	/**
	 * @return the number of faces (triangles) stored in this mesh
	 */
	virtual int getNumFaces ( void ) = 0;

	/**
	 * @return the number of vertices stored in this mesh
	 */
	virtual int getNumVertices ( void ) = 0;

	/**
	 * Adds a face of the mesh to the list of faces
	 *
	 * @param face an array of 3 ints that index into the vertex array
	 *             for this face
	 */
	virtual int addFace ( Face* face ) = 0;

	int addFace ( Face &face )
	{
		return addFace ( &face );
	}

	/** 
	 * @return the face at the given index
	 */
	virtual Face* getFace ( int i ) = 0;

	/** 
	 * @return the number of componets (float values) per vertex
	 */
	int getComponentsPerVertex ( void )
	{
		return numComponents;
	}

	void setComponentsPerVertex ( int num )
	{
		numComponents = num;
	}

	/**
	 * Adds a vertex to the list of vertices
	 *
	 * @param vertex a float array of size numComponents to copy in
	 */
	virtual int addVertex ( Vertex *vertex ) = 0;

	/**
	 * @return the ith vertex
	 */
	virtual Vertex* getVertex ( int i ) = 0;

	void smooth ( int iterations );
};

#endif
