#include <iostream>
#include <fstream>
#include <string>
#include <assert.h>
#include <cstdlib>
#include <boost/filesystem.hpp>
using namespace boost::filesystem;

#include "file_inline.h"
#include "pointset.h"
#include "PlyLoader.h"

using namespace std;

void readPlyPrimal(string filename, PointSet &pointset)
{
	Mesh *mesh = PlyLoader::loadFile((char*)filename.c_str());
	const int vert_num = mesh->getNumVertices();
	const int face_num = mesh->getNumFaces();

	// load original positions
	vector<vect3d> orig;
	orig.resize(vert_num);
	for (int i = 0; i < vert_num; ++i)
	{
		Vertex *tv = mesh->getVertex(i);
		orig[i][0] = tv->getX();
		orig[i][1] = tv->getY();
		orig[i][2] = tv->getZ();
	}

	// smooth
	mesh->smooth(10);

	// load positions
	vector<vect3d> pts;
	pts.resize(vert_num);
	for (int i = 0; i < vert_num; ++i)
	{
		Vertex *tv = mesh->getVertex(i);
		pts[i][0] = tv->getX();
		pts[i][1] = tv->getY();
		pts[i][2] = tv->getZ();
	}

	// zero normals
	vector<vect3d> norms;
	norms.resize(vert_num);
	for (int i = 0; i < vert_num; ++i)
	{
		norms[i](0,0,0);
	}

	// build normals
	for (int i = 0; i < face_num; ++i)
	{
		Face *face = mesh->getFace(i);
		int *data = face->getData();
		assert(face->getLength() == 3);

		const vect3d &v0 = pts[data[0]];
		const vect3d &v1 = pts[data[1]];
		const vect3d &v2 = pts[data[2]];

		const vect3d n = (v1 - v0) % (v2 - v0);

		for (int k = 0; k < 3; ++k)
			norms[data[k]] += n;
	}

	// normalize normals
	for (int i = 0; i < vert_num; ++i)
	{
		vect3d &n = norms[i];
		double len2inv = 1.0 / n.length2();
		if (isfinite(len2inv))
		{
			n *= sqrt(len2inv);
			
			OrientedPoint pt;
			vec3Copy(pt.norm, n);
			vec3Copy(pt.pos, orig[i]);

			pointset.points.push_back(pt);
		}
	}

	// free memory
	delete mesh;
}

void readPlyDual(string filename, PointSet &pointset)
{
	Mesh *mesh = PlyLoader::loadFile((char*)filename.c_str());
	const int vert_num = mesh->getNumVertices();
	const int face_num = mesh->getNumFaces();

	// load original positions
	vector<vect3d> orig;
	orig.resize(vert_num);
	for (int i = 0; i < vert_num; ++i)
	{
		Vertex *tv = mesh->getVertex(i);
		orig[i][0] = tv->getX();
		orig[i][1] = tv->getY();
		orig[i][2] = tv->getZ();
	}
	
	// smooth the mesh to improve normals
	mesh->smooth(10);

	// load positions
	vector<vect3d> pts;
	pts.resize(vert_num);
	for (int i = 0; i < vert_num; ++i)
	{
		Vertex *tv = mesh->getVertex(i);
		pts[i][0] = tv->getX();
		pts[i][1] = tv->getY();
		pts[i][2] = tv->getZ();
	}

	// build normals
	for (int i = 0; i < face_num; ++i)
	{
		Face *face = mesh->getFace(i);
		int *data = face->getData();
		assert(face->getLength() == 3);

		const vect3d &v0 = pts[data[0]];
		const vect3d &v1 = pts[data[1]];
		const vect3d &v2 = pts[data[2]];

		const vect3d &origv0 = orig[data[0]];
		const vect3d &origv1 = orig[data[1]];
		const vect3d &origv2 = orig[data[2]];

		OrientedPoint pt;
		vect3d n = ~((v1 - v0) % (v2 - v0));

		if (isfinite(n[0]) && isfinite(n[1]) && isfinite(n[2]))
		{
//			vect3d p = (origv0 + origv1 + origv2) * (1.0/3.0);
			vect3d p = (v0 + v1 + v2) * (1.0/3.0);

			vec3Copy(pt.norm, n);
			vec3Copy(pt.pos, p);

			pointset.points.push_back(pt);
		}
	}

	// free memory
	delete mesh;
}

void readPly(string filename, PointSet &pointset)
{
	readPlyPrimal(filename, pointset);
	//readPlyDual(filename, pointset);
}

void readModel(string dirname)
{
	int point_num = 0;
	vect3f mine, maxe;

	// write the temporary header
	ofstream file((dirname + ".pts").c_str(), ios_base::binary);

	writeInline(file, mine);
	writeInline(file, maxe);
	writeInline(file, point_num);

	// get names of files
	vector<string> names;
	directory_iterator end_itr; // default construction yields past-the-end
	for (directory_iterator itr(dirname); itr != end_itr; ++itr)
	{
		if (!is_directory(itr->status()))
		{
			string fn = itr->path().leaf();
			names.push_back(fn);
		}
	}

	// read ply files
	bool first = true;
	for (int i = 0; i < names.size(); i++)
	{
		// read the section
		PointSet pointset;
		string filename = dirname + "/" + names[i];
		readPly(filename, pointset); // read each section in directory and create a pts file for it
		printf("read %zu points from %s\n", pointset.points.size(), filename.c_str());
	
		// write the pointset out to disk
		writeInline(file, &pointset.points[0], pointset.points.size());

		// update variables
		findExtents(pointset.points, pointset.mine, pointset.maxe);

		point_num += pointset.points.size();

		if (first)
		{
			first = false;

			mine = pointset.mine;
			maxe = pointset.maxe;
		}
		else
		{
			for (int i = 0; i < 3; i++)
			{
				if (pointset.mine[i] < mine[i])
					mine[i] = pointset.mine[i];

				if (pointset.maxe[i] > maxe[i])
					maxe[i] = pointset.maxe[i];
			}
		}
	}

	// print some info
	printf("---------------------------------\n");
	printf("%d points total\n", point_num);

	// overwrite the header now that the info is known
	file.seekp(0, ios_base::beg);
 
	squareExtents(mine, maxe);

	writeInline(file, mine);
	writeInline(file, maxe);
	writeInline(file, point_num);
	
	file.close();
}

int main(int argc, char **argv)
{
	for (int i = 1; i < argc; i++)	
	{
		readModel(argv[i]);
	}
}
