#include "vect.h"
#include "parse.h"
#include <string>
#include <vector>
#include <fstream>
#include <algorithm>
#include <time.h>

using namespace std;

vect3f mine, maxe; // min and max extents of the bounding box
int axis = 2; // 0 = x, 1 = y, 2 = z => points will be aligned with longest axis in z direction, so axis = 2 will result in no transformation
int pts_per_chunk = 75*1000*1000;
volatile __int32 pts_num = 0;
const int buffer_size = 8*1024*1024;

struct OrientedPoint
{
	vect3f norm, pos;
};

struct ChunkDisk
{
	ifstream *file;
	string filename;
	int pts_total;
	int pts_read;
	OrientedPoint current_pt;
	char *buffer;

	void open(string filename_)
	{
		filename = filename_;
		file = new ifstream(filename.c_str(), ios::binary);
		pts_read = 0;

		buffer = new char [buffer_size];
		file->rdbuf()->pubsetbuf(buffer, buffer_size); // set buffer to be nice and large
		
		if (file->fail())
			printf("FAIL ON OPEN %s!\n", filename.c_str());

		vect3f tmp_v;
		readInline(*file, tmp_v);
		readInline(*file, tmp_v);
		readInline(*file, pts_total);
		read();
	}

	void close()
	{
		file->close();
		delete file;
		delete buffer;
	}

	bool done()
	{
		return pts_read >= pts_total;
	}

	void read()
	{
		OrientedPoint pt_before = current_pt;
		readInline(*file, current_pt);
		pts_read++;

		if (pt_before.pos[0] == current_pt.pos[0] &&
			pt_before.pos[1] == current_pt.pos[1] &&
			pt_before.pos[2] == current_pt.pos[2] &&
			pt_before.norm[0] == current_pt.norm[0] &&
			pt_before.norm[1] == current_pt.norm[1] &&
			pt_before.norm[2] == current_pt.norm[2])
			printf("REPEAT IN FILE %s (%f, %f, %f)\n", filename.c_str(), current_pt.pos[0], current_pt.pos[1], current_pt.pos[2]);

		if (file->fail() || file->eof())
			printf("FAIL ON READ %s!\n", filename.c_str());
	}
};

vect3f axisTransform(const vect3f &pt)
{
	if (axis == 1)
	{
		vect3f ret;
		ret.v[0] = pt.v[0];
		ret.v[1] = -pt.v[2];
		ret.v[2] = pt.v[1];
		return ret;
	}
	else if (axis == 0)
	{
		vect3f ret;
		ret.v[0] = -pt.v[2];
		ret.v[1] = pt.v[1];
		ret.v[2] = pt.v[0];
		return ret;
	}
	else
		return pt;
}

void chooseAxis()
{
	vect3f d = maxe - mine;

	if (d[0] >= d[1] && d[0] >= d[2])
		axis = 0;
	else if (d[1] >= d[0] && d[1] >= d[2])
		axis = 1;
	else if (d[2] >= d[1] && d[2] >= d[0])
		axis = 2;

	mine = axisTransform(mine);
	maxe = axisTransform(maxe);

	for (int i = 0; i < 3; i++)
	{
		if (maxe[i] < mine[i])
			swap(maxe[i], mine[i]);
	}
}

string chunkFileName(int chunk_index)
{
	char buf[256];
	sprintf(buf, "tmp_chunk_%05d.dat", chunk_index);
	return string(buf);
}

bool zOrdering(const OrientedPoint &a, const OrientedPoint &b)
{
	return a.pos.v[2] < b.pos.v[2];
}

void sortChunks(string infilename)
{
	ifstream file(infilename.c_str(), ios::binary);
	if (file.fail())
		return;

	// read the header
	readInline(file, mine);
	readInline(file, maxe);
	readInline(file, pts_num);

	printf ( "mine = %f,%f,%f  maxe = %f,%f,%f\n", mine [ 0 ], mine [ 1 ], mine [ 2 ], maxe [ 0 ], maxe [ 1 ], maxe [ 2 ] );
	for ( int i = 0; i < 3; i++ )
	{
		if ( mine [ i ] > maxe [ i ] )
		{
			printf ( "%d wrong\n", i );
		}
	}

	// chose the axis that data will be aligned along
	chooseAxis();

	printf ( "mine = %f,%f,%f  maxe = %f,%f,%f\n", mine [ 0 ], mine [ 1 ], mine [ 2 ], maxe [ 0 ], maxe [ 1 ], maxe [ 2 ] );

	// sort each chunk and write it to disk
	vector<OrientedPoint> chunk;
	int pts_remaining = pts_num;
	int chunk_index = 0;
	int chunk_num = (pts_num / pts_per_chunk) + (pts_num % pts_per_chunk ? 1 : 0);
		
	printf("There will be %d chunks containing max of %d points each.\n", chunk_num, pts_per_chunk);

	for (;;)
	{
		int chunk_size = pts_per_chunk;

		if (pts_remaining < pts_per_chunk)
			chunk_size = pts_remaining;

		// read a chunk
		chunk.resize(chunk_size);
		readInline(file, &chunk[0], chunk_size);

		// align chunk with z
		for (int i = 0; i < chunk.size(); i++)
		{
			chunk[i].norm = axisTransform(chunk[i].norm);
			chunk[i].pos = axisTransform(chunk[i].pos);
		}

		// sort the chunk
		sort(chunk.begin(), chunk.end(), zOrdering);

		// write the chunk
		string chunkfilename = chunkFileName(chunk_index);
		printf("writing %s\n", chunkfilename.c_str());

		ofstream outfile(chunkfilename.c_str(), ios::binary);
		writeInline(outfile, mine);
		writeInline(outfile, maxe);
		writeInline(outfile, chunk);
		outfile.close();

		// counters
		pts_remaining -= chunk_size;
		chunk_index++;

		if (pts_remaining <= 0)
			break;
	}

	file.close();
}

void mergeChunks(string outfilename)
{
	// combine all of the chunks files into one sorted file
	printf("Combining chunks\n");

	ofstream outfile(outfilename.c_str(), ios::binary);

	//int chunk_num = 4; // DEBUG
	//pts_num = pts_per_chunk * chunk_num; // DEBUG

	// write the header
	writeInline(outfile, mine);
	writeInline(outfile, maxe);
	writeInline(outfile, pts_num);

	int chunk_num = (pts_num / pts_per_chunk) + (pts_num % pts_per_chunk ? 1 : 0);
	vector<ChunkDisk> chunks(chunk_num);
	for (int i = 0; i < chunks.size(); i++)
		chunks[i].open(chunkFileName(i));

	int counter = 0;
	while (chunks.size())
	{
		// find the lowest z value point
		int lowest_index = 0;
		for (int i = 1; i < chunks.size(); i++)
		{
			if (chunks[i].current_pt.pos.v[2] < chunks[lowest_index].current_pt.pos.v[2])
				lowest_index = i;
		}

		if ( counter % 100000 == 0 )
		{
			printf ( "%f,%f,%f: %s\n", chunks [ lowest_index ].current_pt.pos [ 0 ], chunks [ lowest_index ].current_pt.pos [ 1 ], chunks [ lowest_index ].current_pt.pos [ 2 ], chunks [ lowest_index ].filename.c_str ( ) );
		}

		// write the point to disk
		writeInline(outfile, chunks[lowest_index].current_pt);

		// read another point or remove the chunk
		if (chunks[lowest_index].done())
		{
			printf("Finished reading chunk %d (%s).\n", lowest_index, chunks[lowest_index].filename.c_str());
			chunks[lowest_index].close();
			swap(chunks[lowest_index], chunks[chunks.size() - 1]);
			chunks.pop_back();
		}
		else
			chunks[lowest_index].read();
		counter++;
	}
}

void sortFile(string infilename, string outfilename)
{
	int start_time = clock();

	sortChunks(infilename);
	mergeChunks(outfilename);

	// delete all of the temporary chunk files
	
	printf("Sorting complete\n");
	printf("Time taken: %fs\n", (clock() - start_time) / float(CLOCKS_PER_SEC));
}

int main(int argc, char **argv)
{
	string infile, outfile;

	if (argc == 2)
		infile = outfile = argv[1];
	else if (argc == 3)
	{
		infile = argv[1];
		outfile = argv[2];
	}
	else
	{
		printf("usage: sortpoints <infile> [outfile]\n");
		return 0;
	}

	sortFile(infile, outfile);
}
